"""Adds useful t0_idx and sample period attributes"""
import logging
from datetime import timedelta
from typing import Union

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("add_t0_idx_and_sample_period_duration")
class AddT0IdxAndSamplePeriodDurationIterDataPipe(IterDataPipe):
    """Add t0_idx and sample_period_duration attributes to datasets for downstream tasks"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        sample_period_duration: timedelta,
        history_duration: timedelta,
    ):
        """
        Adds two attributes, t0_idx, and sample_period_duration for downstream datapipes to use

        Args:
            source_datapipe: Datapipe emitting a Xarray DataSet or DataArray
            sample_period_duration: Time between samples
            history_duration: Amount of history in each example
        """
        self.source_datapipe = source_datapipe
        self.sample_period_duration = sample_period_duration
        self.history_duration = history_duration
        self.t0_idx = int(self.history_duration / self.sample_period_duration)

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Adds the two attributes to the xarray objects and returns them"""
        for xr_data in self.source_datapipe:
            logger.debug("Adding t0 and sample_period_duration to xarray")
            xr_data.attrs["t0_idx"] = self.t0_idx
            xr_data.attrs["sample_period_duration"] = self.sample_period_duration
            yield xr_data



"""Various conversion datapipes"""
from .coordinates import ConvertGeostationaryToLatLonIterDataPipe as ConvertGeostationaryToLatLon
from .coordinates import ConvertLatLonToOSGBIterDataPipe as ConvertLatLonToOSGB
from .coordinates import ConvertOSGBToLatLonIterDataPipe as ConvertOSGBToLatLon
from .gsp import ConvertGSPToNumpyBatchIterDataPipe as ConvertGSPToNumpyBatch
from .nwp import ConvertNWPToNumpyBatchIterDataPipe as ConvertNWPToNumpyBatch
from .pv import ConvertPVToNumpyBatchIterDataPipe as ConvertPVToNumpyBatch
from .satellite import ConvertSatelliteToNumpyBatchIterDataPipe as ConvertSatelliteToNumpyBatch



from ocf_datapipes.select import DropGSP, LocationPicker
from ocf_datapipes.transform.xarray import ConvertToNWPTargetTime, PreProcessMetNet


def test_metnet_preprocess(sat_datapipe, gsp_datapipe):
    gsp_datapipe = DropGSP(gsp_datapipe, gsps_to_keep=[0])
    gsp_datapipe = LocationPicker(gsp_datapipe)
    datapipe = PreProcessMetNet(
        [sat_datapipe],
        location_datapipe=gsp_datapipe,
        center_width=100_000,
        center_height=100_000,
        context_height=1_000_000,
        context_width=1_000_000,
        output_width_pixels=100,
        output_height_pixels=100,
    )
    data = next(iter(datapipe))
    print(data)
    print(data.shape)



from ocf_datapipes.load import OpenPVFromNetCDF


def test_remove_nans_all_pv_systems(pv_parquet_file):

    pv_datapipe = OpenPVFromNetCDF(
        pv_power_filename=pv_parquet_file,
        pv_metadata_filename="tests/data/pv/passiv/UK_PV_metadata.csv",
    )

    pv_datapipe = pv_datapipe.remove_nans()
    pv_datapipe = pv_datapipe.check_nans()

    data = next(iter(pv_datapipe))
    assert data is not None



"""Geospatial functions"""
from datetime import datetime
from numbers import Number
from typing import Union

import numpy as np
import pandas as pd
import pvlib
import pyproj

# OSGB is also called "OSGB 1936 / British National Grid -- United
# Kingdom Ordnance Survey".  OSGB is used in many UK electricity
# system maps, and is used by the UK Met Office UKV model.  OSGB is a
# Transverse Mercator projection, using 'easting' and 'northing'
# coordinates which are in meters.  See https://epsg.io/27700
OSGB36 = 27700

# WGS84 is short for "World Geodetic System 1984", used in GPS. Uses
# latitude and longitude.
WGS84 = 4326


_osgb_to_lat_lon = pyproj.Transformer.from_crs(crs_from=OSGB36, crs_to=WGS84).transform
_lat_lon_to_osgb = pyproj.Transformer.from_crs(crs_from=WGS84, crs_to=OSGB36).transform


def osgb_to_lat_lon(
    x: Union[Number, np.ndarray], y: Union[Number, np.ndarray]
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change OSGB coordinates to lat, lon.

    Args:
        x: osgb east-west
        y: osgb north-south
    Return: 2-tuple of latitude (north-south), longitude (east-west).
    """
    return _osgb_to_lat_lon(xx=x, yy=y)


def lat_lon_to_osgb(
    latitude: Union[Number, np.ndarray], longitude: Union[Number, np.ndarray]
) -> tuple[Union[Number, np.ndarray], Union[Number, np.ndarray]]:
    """Change lat, lon coordinates to OSGB.

    Return: 2-tuple of OSGB x, y.
    """
    return _lat_lon_to_osgb(xx=latitude, yy=longitude)


def calculate_azimuth_and_elevation_angle(
    latitude: float, longitude: float, datestamps: list[datetime]
) -> pd.DataFrame:
    """
    Calculation the azimuth angle, and the elevation angle for several datetamps.

    But for one specific osgb location

    More details see:
    https://www.celestis.com/resources/faq/what-are-the-azimuth-and-elevation-of-a-satellite/

    Args:
        latitude: latitude of the pv site
        longitude: longitude of the pv site
        datestamps: list of datestamps to calculate the sun angles. i.e the sun moves from east to
            west in the day.

    Returns: Pandas data frame with the index the same as 'datestamps', with columns of
    "elevation" and "azimuth" that have been calculate.

    """
    # get the solor position
    solpos = pvlib.solarposition.get_solarposition(datestamps, latitude, longitude)

    # extract the information we want
    return solpos[["elevation", "azimuth"]]


def load_geostationary_area_definition_and_transform_osgb(xr_data):
    """
    Loads geostationary area and transformation from OSGB to geostationaery

    Args:
        xr_data: Xarray object with geostationary area

    Returns:
        The transform
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    area_definition_yaml = xr_data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    osgb_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=OSGB36, crs_to=geostationary_crs
    ).transform
    return osgb_to_geostationary


def load_geostationary_area_definition_and_transform_latlon(xr_data):
    """
    Loads geostationary area and transformation from Latlon to geostationaery

    Args:
        xr_data: Xarray object with geostationary area

    Returns:
        The transform
    """
    # Only load these if using geostationary projection
    import pyproj
    import pyresample

    area_definition_yaml = xr_data.attrs["area"]
    geostationary_area_definition = pyresample.area_config.load_area_from_string(
        area_definition_yaml
    )
    geostationary_crs = geostationary_area_definition.crs
    latlon_to_geostationary = pyproj.Transformer.from_crs(
        crs_from=WGS84, crs_to=geostationary_crs
    ).transform
    return latlon_to_geostationary



"""Datapipe and utils to load PV data from NetCDF for training"""
import logging
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.load.pv.utils import (
    intersection_of_pv_system_ids,
    put_pv_data_into_an_xr_dataarray,
)
from ocf_datapipes.utils.geospatial import lat_lon_to_osgb

_log = logging.getLogger(__name__)


@functional_datapipe("open_pv_netcdf")
class OpenPVFromNetCDFIterDataPipe(IterDataPipe):
    """Datapipe to load NetCDF"""

    def __init__(
        self,
        pv_power_filename: Union[str, Path],
        pv_metadata_filename: Union[str, Path],
        start_datetime: Optional[datetime] = None,
        end_datetime: Optional[datetime] = None,
    ):
        """
        Datapipe to load PV from NetCDF

        Args:
            pv_power_filename: Filename of the power file
            pv_metadata_filename: Filename of the metadata file
            start_datetime: start datetime that the dataset is limited to
            end_datetime: end datetime that the dataset is limited to
        """
        super().__init__()
        self.pv_power_filename = pv_power_filename
        self.pv_metadata_filename = pv_metadata_filename
        self.start_dateime = start_datetime
        self.end_datetime = end_datetime

    def __iter__(self):
        data: xr.DataArray = load_everything_into_ram(
            self.pv_power_filename,
            self.pv_metadata_filename,
            start_dateime=self.start_dateime,
            end_datetime=self.end_datetime,
        )
        while True:
            yield data


def load_everything_into_ram(
    pv_power_filename,
    pv_metadata_filename,
    start_dateime: Optional[datetime] = None,
    end_datetime: Optional[datetime] = None,
) -> xr.DataArray:
    """Open AND load PV data into RAM."""

    # load metadata
    pv_metadata = _load_pv_metadata(pv_metadata_filename)

    # Load pd.DataFrame of power and pd.Series of capacities:
    (
        pv_power_watts,
        pv_capacity_watt_power,
        pv_system_row_number,
    ) = _load_pv_power_watts_and_capacity_watt_power(
        pv_power_filename,
        start_date=start_dateime,
        end_date=end_datetime,
    )
    # Ensure pv_metadata, pv_power_watts, and pv_capacity_watt_power all have the same set of
    # PV system IDs, in the same order:
    pv_metadata, pv_power_watts = intersection_of_pv_system_ids(pv_metadata, pv_power_watts)
    pv_capacity_watt_power = pv_capacity_watt_power.loc[pv_power_watts.columns]
    pv_system_row_number = pv_system_row_number.loc[pv_power_watts.columns]

    data_in_ram = put_pv_data_into_an_xr_dataarray(
        pv_power_watts=pv_power_watts,
        y_osgb=pv_metadata.y_osgb.astype(np.float32),
        x_osgb=pv_metadata.x_osgb.astype(np.float32),
        capacity_watt_power=pv_capacity_watt_power,
        pv_system_row_number=pv_system_row_number,
    )

    # Sanity checks:
    time_utc = pd.DatetimeIndex(data_in_ram.time_utc)
    assert time_utc.is_monotonic_increasing
    assert time_utc.is_unique

    return data_in_ram


def _load_pv_power_watts_and_capacity_watt_power(
    filename: Union[str, Path],
    start_date: Optional[datetime] = None,
    end_date: Optional[datetime] = None,
) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
    """Return pv_power_watts, pv_capacity_watt_power, pv_system_row_number.

    The capacities and pv_system_row_number are computed across the *entire* dataset,
    and so is independent of the `start_date` and `end_date`. This ensures the PV system
    row number and capacities stay constant across training and validation.
    """

    _log.info(f"Loading solar PV power data from {filename} from {start_date=} to {end_date=}.")

    # Load data in a way that will work in the cloud and locally:
    if ".parquet" in str(filename):
        _log.debug(f"Loading PV parquet file {filename}")
        pv_power_df = pd.read_parquet(filename, engine="fastparquet")
        _log.debug("Loading PV parquet file: done")
        pv_power_df["generation_w"] = pv_power_df["generation_wh"] * 12
        if end_date is not None:
            pv_power_df = pv_power_df[pv_power_df["timestamp"] < end_date]
        if start_date is not None:
            pv_power_df = pv_power_df[pv_power_df["timestamp"] >= start_date]

        # pivot on ss_id
        _log.debug("Pivoting PV data")
        pv_power_watts = pv_power_df.pivot(
            index="timestamp", columns="ss_id", values="generation_w"
        )
        _log.debug("Pivoting PV data: done")
        pv_capacity_watt_power = pv_power_watts.max().astype(np.float32)

    else:
        with fsspec.open(filename, mode="rb") as file:
            pv_power_ds = xr.open_dataset(file, engine="h5netcdf")
            pv_capacity_watt_power = pv_power_ds.max().to_pandas().astype(np.float32)
            pv_power_watts = pv_power_ds.sel(datetime=slice(start_date, end_date)).to_dataframe()
            pv_power_watts = pv_power_watts.astype(np.float32)

            del pv_power_ds

        if "passiv" not in str(filename):
            _log.warning("Converting timezone. ARE YOU SURE THAT'S WHAT YOU WANT TO DO?")
            try:
                pv_power_watts = (
                    pv_power_watts.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)
                )
            except Exception as e:
                _log.warning(
                    "Could not convert timezone from London to UTC. "
                    "Going to try and carry on anyway"
                )
                _log.warning(e)

    pv_capacity_watt_power.index = [np.int32(col) for col in pv_capacity_watt_power.index]
    pv_power_watts.columns = pv_power_watts.columns.astype(np.int64)

    # Create pv_system_row_number. We use the index of
    # `pv_capacity_watt_power` because that includes
    # the PV system IDs for the entire dataset (independent of `start_date` and `end_date`).
    # We use `float32` for the ID because we use NaN to indicate a missing PV system,
    # or that this whole example doesn't include PV.
    all_pv_system_ids = pv_capacity_watt_power.index
    pv_system_row_number = np.arange(start=0, stop=len(all_pv_system_ids), dtype=np.float32)
    pv_system_row_number = pd.Series(pv_system_row_number, index=all_pv_system_ids)

    _log.info(
        "After loading:"
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )

    pv_power_watts = pv_power_watts.clip(lower=0, upper=5e7)
    pv_power_watts = _drop_pv_systems_which_produce_overnight(pv_power_watts)

    # Resample to 5-minutely and interpolate up to 15 minutes ahead.
    # TODO: Issue #74: Give users the option to NOT resample (because Perceiver IO
    # doesn't need all the data to be perfectly aligned).
    pv_power_watts = pv_power_watts.resample("5T").interpolate(method="time", limit=3)
    pv_power_watts.dropna(axis="index", how="all", inplace=True)
    pv_power_watts.dropna(axis="columns", how="all", inplace=True)

    # Drop any PV systems whose PV capacity is too low:
    PV_CAPACITY_THRESHOLD_W = 100
    pv_systems_to_drop = pv_capacity_watt_power.index[
        pv_capacity_watt_power <= PV_CAPACITY_THRESHOLD_W
    ]
    pv_systems_to_drop = pv_systems_to_drop.intersection(pv_power_watts.columns)
    _log.info(
        f"Dropping {len(pv_systems_to_drop)} PV systems because their max power is less than"
        f" {PV_CAPACITY_THRESHOLD_W}"
    )
    pv_power_watts.drop(columns=pv_systems_to_drop, inplace=True)

    # Ensure that capacity and pv_system_row_num use the same PV system IDs as the power DF:
    pv_system_ids = pv_power_watts.columns
    pv_capacity_watt_power = pv_capacity_watt_power.loc[pv_system_ids]
    pv_system_row_number = pv_system_row_number.loc[pv_system_ids]

    _log.info(
        "After filtering & resampling to 5 minutes:"
        f" pv_power = {pv_power_watts.values.nbytes / 1e6:,.1f} MBytes."
        f" {len(pv_power_watts)} PV power datetimes."
        f" {len(pv_power_watts.columns)} PV power PV system IDs."
    )

    # Sanity checks:
    assert not pv_power_watts.columns.duplicated().any()
    assert not pv_power_watts.index.duplicated().any()
    assert np.isfinite(pv_capacity_watt_power).all()
    assert (pv_capacity_watt_power >= 0).all()
    assert np.isfinite(pv_system_row_number).all()
    assert np.array_equal(pv_power_watts.columns, pv_capacity_watt_power.index)
    return pv_power_watts, pv_capacity_watt_power, pv_system_row_number


# Adapted from nowcasting_dataset.data_sources.pv.pv_data_source
def _load_pv_metadata(filename: str) -> pd.DataFrame:
    """Return pd.DataFrame of PV metadata.

    Shape of the returned pd.DataFrame for Passiv PV data:
        Index: ss_id (Sheffield Solar ID)
        Columns: llsoacd, orientation, tilt, kwp, operational_at,
            latitude, longitude, system_id, x_osgb, y_osgb
    """
    _log.info(f"Loading PV metadata from {filename}")

    if "passiv" in str(filename):
        index_col = "ss_id"
    else:
        index_col = "system_id"

    pv_metadata = pd.read_csv(filename, index_col=index_col)

    if "Unnamed: 0" in pv_metadata.columns:
        pv_metadata.drop(columns="Unnamed: 0", inplace=True)

    _log.info(f"Found {len(pv_metadata)} PV systems in {filename}")

    # drop any systems with no lon or lat:
    pv_metadata.dropna(subset=["longitude", "latitude"], how="any", inplace=True)

    _log.debug(f"Found {len(pv_metadata)} PV systems with locations")

    pv_metadata["x_osgb"], pv_metadata["y_osgb"] = lat_lon_to_osgb(
        latitude=pv_metadata["latitude"], longitude=pv_metadata["longitude"]
    )

    return pv_metadata


def _drop_pv_systems_which_produce_overnight(pv_power_watts: pd.DataFrame) -> pd.DataFrame:
    """Drop systems which produce power over night.

    Args:
        pv_power_watts: Un-normalised.
    """
    # TODO: Of these bad systems, 24647, 42656, 42807, 43081, 51247, 59919
    # might have some salvagable data?
    NIGHT_YIELD_THRESHOLD = 0.4
    night_hours = [22, 23, 0, 1, 2]
    pv_power_normalised = pv_power_watts / pv_power_watts.max()
    night_mask = pv_power_normalised.index.hour.isin(night_hours)
    pv_power_at_night_normalised = pv_power_normalised.loc[night_mask]
    pv_above_threshold_at_night = (pv_power_at_night_normalised > NIGHT_YIELD_THRESHOLD).any()
    bad_systems = pv_power_normalised.columns[pv_above_threshold_at_night]
    _log.info(f"{len(bad_systems)} bad PV systems found and removed!")
    return pv_power_watts.drop(columns=bad_systems)



"""Datapipes to add Sun position to NumpyBatch"""
import warnings

import numpy as np
import pandas as pd
import pvlib
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey
from ocf_datapipes.utils.geospatial import osgb_to_lat_lon

ELEVATION_MEAN = 37.4
ELEVATION_STD = 12.7
AZIMUTH_MEAN = 177.7
AZIMUTH_STD = 41.7


@functional_datapipe("add_sun_position")
class AddSunPositionIterDataPipe(IterDataPipe):
    """Adds the sun position to the NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe, modality_name: str):
        """
        Adds the sun position to the NumpyBatch

        Args:
            source_datapipe: Datapipe of NumpyBatch
            modality_name: modality to add the sun position for
        """
        self.source_datapipe = source_datapipe
        self.modality_name = modality_name
        assert self.modality_name in [
            "hrvsatellite",
            "gsp",
            "gsp_5_min",
            "pv",
            "nwp_target_time",
            "satellite",
        ], f"Cant add sun position on {self.modality_name}"

    def __iter__(self):
        for np_batch in self.source_datapipe:
            # TODO Make work with Lat/Lons instead
            if self.modality_name == "hrvsatellite":
                y_osgb = np_batch[BatchKey.hrvsatellite_y_osgb]  # example, y, x
                x_osgb = np_batch[BatchKey.hrvsatellite_x_osgb]  # example, y, x
                time_utc = np_batch[BatchKey.hrvsatellite_time_utc]  # example, time

                # Get the time and position for the centre of the t0 frame:
                y_centre_idx = int(y_osgb.shape[1] // 2)
                x_centre_idx = int(y_osgb.shape[2] // 2)
                y_osgb_centre = y_osgb[:, y_centre_idx, x_centre_idx]  # Shape: (example,)
                x_osgb_centre = x_osgb[:, y_centre_idx, x_centre_idx]  # Shape: (example,)
            elif self.modality_name == "pv":
                pv_y_osgb = np_batch[BatchKey.pv_y_osgb]
                pv_x_osgb = np_batch[BatchKey.pv_x_osgb]
                time_utc = np_batch[BatchKey.pv_time_utc]
                # Sometimes, the PV coords can all be NaNs if there are no PV systems
                # for that datetime and location. e.g. in northern Scotland!
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action="ignore", category=RuntimeWarning, message="Mean of empty slice"
                    )
                    y_osgb_centre = np.nanmean(pv_y_osgb, axis=1)
                    x_osgb_centre = np.nanmean(pv_x_osgb, axis=1)
            elif self.modality_name == "nwp_target_time":
                y_osgb_centre = np_batch[BatchKey.nwp_y_osgb].mean(axis=-1)
                x_osgb_centre = np_batch[BatchKey.nwp_x_osgb].mean(axis=-1)
                time_utc = np_batch[BatchKey.nwp_target_time_utc]
            else:  # gsp and gsp_5_min:
                y_osgb_centre = np_batch[BatchKey.gsp_y_osgb]
                x_osgb_centre = np_batch[BatchKey.gsp_x_osgb]
                time_utc = np_batch[BatchKey[self.modality_name + "_time_utc"]]

                # Convert to the units that pvlib expects: lat, lon.
            lats, lons = osgb_to_lat_lon(x=x_osgb_centre, y=y_osgb_centre)

            # Loop round each example to get the Sun's elevation and azimuth:
            azimuth = np.full_like(time_utc, fill_value=np.NaN).astype(np.float32)
            elevation = np.full_like(time_utc, fill_value=np.NaN).astype(np.float32)
            must_be_finite = True
            for example_idx, (lat, lon) in enumerate(zip(lats, lons)):
                if not np.isfinite([lat, lon]).all():
                    assert (
                        self.modality_name == "pv"
                    ), f"{self.modality_name} lat and lon must be finite! But {lat=} {lon=}!"
                    # This is PV data, for a location which has no PV systems.
                    must_be_finite = False
                    continue  # Leave azimuth and elevation as NaN for this example_idx.
                dt = pd.to_datetime(time_utc[example_idx], unit="s")
                dt = pd.DatetimeIndex(dt)  # pvlib expects a `pd.DatetimeIndex`.
                solpos = pvlib.solarposition.get_solarposition(
                    time=dt,
                    latitude=lat,
                    longitude=lon,
                    # Which `method` to use?
                    # pyephem seemed to be a good mix between speed and ease but causes segfaults!
                    # nrel_numba doesn't work when using multiple worker processes.
                    # nrel_c is probably fastest but requires C code to be manually compiled:
                    # https://midcdmz.nrel.gov/spa/
                )
                azimuth[example_idx] = solpos["azimuth"]
                elevation[example_idx] = solpos["elevation"]

            # Normalise.
            azimuth = (azimuth - AZIMUTH_MEAN) / AZIMUTH_STD
            elevation = (elevation - ELEVATION_MEAN) / ELEVATION_STD

            # Check
            if must_be_finite:
                assert np.isfinite(azimuth).all()
                assert np.isfinite(elevation).all()

            # Store.
            azimuth_batch_key = BatchKey[self.modality_name + "_solar_azimuth"]
            elevation_batch_key = BatchKey[self.modality_name + "_solar_elevation"]
            np_batch[azimuth_batch_key] = azimuth
            np_batch[elevation_batch_key] = elevation

            yield np_batch



"""Datapipe to add topographic data to NumpyBatch"""
import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe, Zipper

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch


@functional_datapipe("add_topographic_data")
class AddTopographicDataIterDataPipe(IterDataPipe):
    """Datapipe to add topographic data to NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe, topo_datapipe: IterDataPipe):
        """
        Datapipe to add topographic data to NumpyBatch

        Args:
            source_datapipe: Datapipe of satellite data
            topo_datapipe: Datapipe of topographic data
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.topo_datapipe = topo_datapipe

    def __iter__(self) -> NumpyBatch:
        """Datapipe to add topographic data to NumpyBatch"""
        for topo, np_batch in Zipper(self.topo_datapipe, self.source_datapipe):
            if BatchKey.hrvsatellite_x_geostationary in np_batch:
                # Recreate an xr.DataArray of the satellite data. This is required so we can
                # use xr.combine_by_coords to align the topo data with the satellite data.
                hrvsatellite_data_array = xr.DataArray(
                    # We're not actually interested in the image. But xarray won't make an
                    # empty DataArray without data in the right shape.
                    # There's nothing special about the x_osgb data. It's just convenient because
                    # it's the right shape!
                    np_batch[BatchKey.hrvsatellite_x_osgb],
                    dims=("example", "y", "x"),
                    coords={
                        "y_geostationary": (
                            ("example", "y"),
                            np_batch[BatchKey.hrvsatellite_y_geostationary],
                        ),
                        "x_geostationary": (
                            ("example", "x"),
                            np_batch[BatchKey.hrvsatellite_x_geostationary],
                        ),
                    },
                )
                hrvsatellite_surface_height = _get_surface_height_for_satellite(
                    surface_height=topo, satellite=hrvsatellite_data_array
                )
                np_batch[BatchKey.hrvsatellite_surface_height] = np.nan_to_num(
                    hrvsatellite_surface_height, nan=0.0
                )
            yield np_batch


def _get_surface_height_for_satellite(
    surface_height: xr.DataArray, satellite: xr.DataArray
) -> np.ndarray:
    """
    Get the surface height for the satellite data

    Args:
        surface_height: Topographic data
        satellite: Satellite data

    Returns:
        The topographic data for each satellite pixel
    """
    num_examples = satellite.shape[0]
    surface_height = surface_height.rename("surface_height")
    surface_height_for_batch = np.full_like(satellite.values, fill_value=np.NaN)
    for example_idx in range(num_examples):
        satellite_example = satellite.isel(example=example_idx)
        msg = "Satellite imagery must start in the top-left!"
        assert satellite_example.y_geostationary[0] > satellite_example.y_geostationary[-1], msg
        assert satellite_example.x_geostationary[0] < satellite_example.x_geostationary[-1], msg
        # This is needed to get the x and y to be dimension coordinates
        # for the combine_by_coords to work
        satellite_example = (
            satellite_example.rename("sat")
            .reset_coords(["x_geostationary", "y_geostationary"])
            .swap_dims({"y": "y_geostationary", "x": "x_geostationary"})
            .to_array()
            .squeeze()
            .reset_coords("variable", drop=True)
            .rename({"y_geostationary": "y", "x_geostationary": "x"})
            .rename("sat")
        )
        surface_height_for_example = surface_height.sel(
            y=slice(
                satellite_example.y[0],
                satellite_example.y[-1],
            ),
            x=slice(
                satellite_example.x[0],
                satellite_example.x[-1],
            ),
        )
        # Align by coordinates. This will result in lots of NaNs in the surface height data:
        aligned = xr.combine_by_coords(
            (surface_height_for_example, satellite_example), join="outer"
        )
        # Fill in the NaNs:
        surface_height_for_example = (
            aligned["surface_height"].ffill("x").ffill("y").bfill("x").bfill("y")
        )

        # Now select exactly the same coordinates from the surface height data as the satellite data
        aligned = xr.combine_by_coords((surface_height_for_example, satellite_example), join="left")

        surface_height_for_batch[example_idx] = aligned["surface_height"].values

    # If we slightly ran off the edge of the topo data then we'll get NaNs.
    # TODO: Enlarge topo data so we never get NaNs!
    surface_height_for_batch = np.nan_to_num(surface_height_for_batch, nan=0)

    return surface_height_for_batch



"""Save out t0 time"""
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch


@functional_datapipe("save_t0_time")
class SaveT0TimeIterDataPipe(IterDataPipe):
    """Save out t0 time"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Save out t0 time

        Args:
            source_datapipe: Datapipe of NumpyBatch objectsK
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        for np_batch in self.source_datapipe:
            pv_t0_idx = np_batch[BatchKey.pv_t0_idx]
            gsp_t0_idx = np_batch[BatchKey.gsp_t0_idx]
            hrvsatellite_t0_idx = np_batch[BatchKey.hrvsatellite_t0_idx]

            np_batch[BatchKey.pv_time_utc_fourier_t0] = np_batch[BatchKey.pv_time_utc_fourier][
                :, pv_t0_idx
            ]
            np_batch[BatchKey.gsp_time_utc_fourier_t0] = np_batch[BatchKey.gsp_time_utc_fourier][
                :, gsp_t0_idx
            ]
            np_batch[BatchKey.hrvsatellite_time_utc_fourier_t0] = np_batch[
                BatchKey.hrvsatellite_time_utc_fourier
            ][:, hrvsatellite_t0_idx]
            yield np_batch



"""Datapipes to add Sun position to NumpyBatch"""

from typing import Optional

from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.config.model import Configuration


@functional_datapipe("add_length")
class AddLengthIterDataPipe(IterDataPipe):
    """Adds length to the datapipe"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        length: Optional[int] = None,
        configuration: Optional[Configuration] = None,
        train_validation_test: Optional[str] = "train",
    ):
        """
        Adds length to the data pipe. This is useful when training a model.

        Either 'length' is set or, 'configuration' and 'train_validation_test' is set

        Args:
            source_datapipe: Datapipe of NumpyBatch
            length: Length of datapipe
            configuration: dataset configuration
            train_validation_test: either 'train', 'validation', or 'test'
        """
        self.source_datapipe = source_datapipe
        self.length = length
        self.configuration = configuration
        self.train_validation_test = train_validation_test

        if self.length is None:
            assert self.configuration is not None
            assert self.train_validation_test in ["train", "validation", "test"]

            if self.train_validation_test == "train":
                self.length = self.configuration.process.n_train_batches
            elif self.train_validation_test == "validation":
                self.length = self.configuration.process.n_validation_batches
            else:
                self.length = self.configuration.process.n_test_batches

    def __len__(self):
        return self.length

    def __iter__(self):
        for np_batch in self.source_datapipe:
            yield np_batch



import numpy as np

from ocf_datapipes.transform.xarray import CreatePVImage


def test_create_pv_image(passiv_datapipe, sat_datapipe):
    pv_image_datapipe = CreatePVImage(passiv_datapipe, sat_datapipe)
    data = next(iter(pv_image_datapipe))
    assert np.max(data) > 0



"""Preprocessing for MetNet-type inputs"""
from typing import List

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe, Zipper

from ocf_datapipes.utils.geospatial import load_geostationary_area_definition_and_transform_osgb


@functional_datapipe("preprocess_metnet")
class PreProcessMetNetIterDataPipe(IterDataPipe):
    """Preprocess set of Xarray datasets similar to MetNet-1"""

    def __init__(
        self,
        source_datapipes: List[IterDataPipe],
        location_datapipe: IterDataPipe,
        context_width: float,
        context_height: float,
        center_width: float,
        center_height: float,
        output_height_pixels: int,
        output_width_pixels: int,
    ):
        """

        Processes set of Xarray datasets similar to MetNet

        In terms of taking all available source datapipes:
        1. selecting the same context area of interest
        2. Creating a center crop of the center_height, center_width
        3. Downsampling the context area of interest to the same shape as the center crop
        4. Stacking those context images on the center crop.

        This would be designed originally for NWP+Satellite+Topographic data sources.
        To add the PV power for lots of sites, the PV power would
        need to be able to be on a grid for the context/center
        crops and then for the downsample

        This also appends Lat/Lon coordinates to the stack,
         and returns a new Numpy array with the stacked data

        TODO Could also add the national PV as a set of Layers, so also GSP input

        Args:
            source_datapipes: Datapipes that emit xarray datasets
                with latitude/longitude coordinates included
            location_datapipe: Datapipe emitting location coordinate for center of example
            context_width: Width of the context area
            context_height: Height of the context area
            center_width: Center width of the area of interest
            center_height: Center height of the area of interest
            output_height_pixels: Output height in pixels
            output_width_pixels: Output width in pixels
        """
        self.source_datapipes = source_datapipes
        self.location_datapipe = location_datapipe
        self.context_width = context_width
        self.context_height = context_height
        self.center_width = center_width
        self.center_height = center_height
        self.output_height_pixels = output_height_pixels
        self.output_width_pixels = output_width_pixels

    def __iter__(self) -> np.ndarray:
        for xr_datas, location in Zipper(Zipper(*self.source_datapipes), self.location_datapipe):
            # TODO Use the Lat/Long coordinates of the center array for the lat/lon stuff
            centers = []
            contexts = []
            for xr_data in xr_datas:
                xr_context: xr.Dataset = _get_spatial_crop(
                    xr_data,
                    location=location,
                    roi_width_meters=self.context_width,
                    roi_height_meters=self.context_height,
                )
                xr_center: xr.Dataset = _get_spatial_crop(
                    xr_data,
                    location=location,
                    roi_width_meters=self.center_width,
                    roi_height_meters=self.center_height,
                )
                # Resamples to the same number of pixels for both center and contexts
                xr_center = _resample_to_pixel_size(
                    xr_center, self.output_height_pixels, self.output_width_pixels
                )
                xr_context = _resample_to_pixel_size(
                    xr_context, self.output_height_pixels, self.output_width_pixels
                )
                xr_center = xr_center.to_numpy()
                xr_context = xr_context.to_numpy()
                if len(xr_center.shape) == 3:  # Need to add channel dimension
                    xr_center = np.expand_dims(xr_center, axis=1)
                    xr_context = np.expand_dims(xr_context, axis=1)
                centers.append(xr_center)
                contexts.append(xr_context)
            # Pad out time dimension to be the same, using the largest one
            # All should have 4 dimensions at this point
            for i in range(len(centers)):
                centers[i] = np.pad(
                    centers[i],
                    pad_width=(
                        (0, np.max([c.shape[0] for c in centers]) - centers[i].shape[0]),
                        (0, 0),
                        (0, 0),
                        (0, 0),
                    ),
                    mode="constant",
                    constant_values=0.0,
                )
                contexts[i] = np.pad(
                    contexts[i],
                    pad_width=(
                        (0, np.max([c.shape[0] for c in contexts]) - contexts[i].shape[0]),
                        (0, 0),
                        (0, 0),
                        (0, 0),
                    ),
                    mode="constant",
                    constant_values=0.0,
                )
            stacked_data = np.concatenate([*centers, *contexts], axis=1)
            yield stacked_data


def _get_spatial_crop(xr_data, location, roi_height_meters: int, roi_width_meters: int):
    # Compute the index for left and right:
    half_height = roi_height_meters // 2
    half_width = roi_width_meters // 2

    left = location.x - half_width
    right = location.x + half_width
    bottom = location.y - half_height
    top = location.y + half_height
    if "x_geostationary" in xr_data.coords:
        # Convert to geostationary edges
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x_geostationary) & (xr_data.x_geostationary <= right)
        y_mask = (xr_data.y_geostationary <= top) & (  # Y is flipped
            bottom <= xr_data.y_geostationary
        )
        selected = xr_data.isel(x_geostationary=x_mask, y_geostationary=y_mask)
    elif "x" in xr_data.coords:
        _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
        left, bottom = _osgb_to_geostationary(xx=left, yy=bottom)
        right, top = _osgb_to_geostationary(xx=right, yy=top)
        x_mask = (left <= xr_data.x) & (xr_data.x <= right)
        y_mask = (xr_data.y <= top) & (bottom <= xr_data.y)  # Y is flipped
        selected = xr_data.isel(x=x_mask, y=y_mask)
    else:
        # Select data in the region of interest:
        x_mask = (left <= xr_data.x_osgb) & (xr_data.x_osgb <= right)
        y_mask = (xr_data.y_osgb <= top) & (bottom <= xr_data.y_osgb)
        selected = xr_data.isel(x_osgb=x_mask, y_osgb=y_mask)

    return selected


def _resample_to_pixel_size(xr_data, height_pixels, width_pixels) -> np.ndarray:
    if "x_geostationary" in xr_data.coords:
        x_coords = xr_data["x_geostationary"].values
        y_coords = xr_data["y_geostationary"].values
    elif "x" in xr_data.coords:
        x_coords = xr_data["x"].values
        y_coords = xr_data["y"].values
    else:
        x_coords = xr_data["x_osgb"].values
        y_coords = xr_data["y_osgb"].values
    # Resample down to the number of pixels wanted
    x_coords = np.linspace(x_coords[0], x_coords[-1], num=width_pixels)
    y_coords = np.linspace(y_coords[0], y_coords[-1], num=height_pixels)
    if "x_geostationary" in xr_data.coords:
        xr_data = xr_data.interp(
            x_geostationary=x_coords, y_geostationary=y_coords, method="linear"
        )
    elif "x" in xr_data.coords:
        xr_data = xr_data.interp(x=x_coords, y=y_coords, method="linear")
    else:
        xr_data = xr_data.interp(x_osgb=x_coords, y_osgb=y_coords, method="linear")
    # Extract just the data now
    return xr_data



""" Data pipes and utils for getting PV data from database"""
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import List, Optional

import numpy as np
import pandas as pd
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models.base import Base_PV
from nowcasting_datamodel.models.pv import (
    PVSystem,
    PVSystemSQL,
    PVYield,
    PVYieldSQL,
    pv_output,
    solar_sheffield_passiv,
)
from nowcasting_datamodel.read.read_pv import get_pv_systems, get_pv_yield
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.load.pv.utils import encode_label, put_pv_data_into_an_xr_dataarray
from ocf_datapipes.utils.geospatial import calculate_azimuth_and_elevation_angle, lat_lon_to_osgb

logger = logging.getLogger(__name__)


@functional_datapipe("open_pv_from_db")
class OpenPVFromDBIterDataPipe(IterDataPipe):
    """Data pipes and utils for getting PV data from database"""

    def __init__(
        self,
        providers: List[str],
        interpolate_minutes: int = 30,
        load_extra_minutes: int = 60,
        history_minutes: int = 30,
    ):
        """
        Datapipe to get PV from database

        Args:
            providers: Providers to use
            interpolate_minutes: How many minutes to interpolate
            load_extra_minutes: How many extra minutes to load
            history_minutes: How many history minutes to use
        """

        super().__init__()

        self.providers = providers
        self.interpolate_minutes = interpolate_minutes
        self.load_extra_minutes = load_extra_minutes
        self.history_minutes = history_minutes

        self.history_duration = pd.Timedelta(self.history_minutes, unit="minutes")

    def __iter__(self):

        pv_metadata = get_metadata_from_database(providers=self.providers)
        pv_power = get_pv_power_from_database(
            providers=self.providers,
            history_duration=self.history_duration,
            interpolate_minutes=self.interpolate_minutes,
            load_extra_minutes=self.load_extra_minutes,
        )

        pv_system_row_number = pd.Series([1] * len(pv_power.columns), index=pv_power.columns)

        # select metadata that is in pv_power
        logger.debug(
            f"There are currently {len(pv_metadata.index)} pv system in the metadata, "
            f"before joining together with power"
        )
        logger.debug(
            f"There are currently {len(pv_power.columns)} pv system in the power data, "
            f"before joining together with metadata"
        )
        pv_metadata = pv_metadata[pv_metadata.index.isin(pv_power.columns)]
        logger.debug(f"There are now {len(pv_metadata.index)} pv system in the metadata")
        logger.debug(f"There are now {len(pv_power.columns)} pv system in the power data")

        data_xr = put_pv_data_into_an_xr_dataarray(
            pv_power_watts=pv_power,  # TODO check this is watts
            y_osgb=pv_metadata.y_osgb.astype(np.float32),
            x_osgb=pv_metadata.x_osgb.astype(np.float32),
            capacity_watt_power=pv_metadata.capacity_watt_power,
            pv_system_row_number=pv_system_row_number,
        )

        logger.info(f"Found {len(data_xr.pv_system_row_number)} PV systems")

        while True:
            yield data_xr


def get_metadata_from_database(providers: List[str] = None) -> pd.DataFrame:
    """
    Get metadata from database

    Returns: pandas data frame with the following columns
        - latitude
        - longitude
        - kwp
        The index is the pv_system_id
    """

    if providers is None:
        providers = [pv_output, solar_sheffield_passiv]

    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_PV)

    pv_system_all_df = []
    for provider in providers:

        logger.debug(f"Get PV systems from database for {provider}")

        with db_connection.get_session() as session:
            # read pv systems
            pv_systems: List[PVSystemSQL] = get_pv_systems(session=session, provider=provider)

            # format locations
            pv_systems_df = pd.DataFrame(
                [(PVSystem.from_orm(pv_system)).__dict__ for pv_system in pv_systems]
            )

        if len(pv_systems_df) == 0:
            pv_systems_df = pd.DataFrame(
                columns=["pv_system_id", "latitude", "longitude", "installed_capacity_kw"]
            )
        else:
            pv_systems_df.index = encode_label(pv_systems_df["pv_system_id"], label=provider)
            pv_systems_df["installed_capacity_kw"] = pv_systems_df["ml_capacity_kw"]
            pv_systems_df = pv_systems_df[["latitude", "longitude", "installed_capacity_kw"]]

        pv_system_all_df.append(pv_systems_df)

    pv_system_all_df = pd.concat(pv_system_all_df)

    logger.debug(f"Found {len(pv_system_all_df)} pv systems")

    # add x_osgb and y_osgb
    x_osgb, y_osgb = lat_lon_to_osgb(
        latitude=pv_system_all_df["latitude"], longitude=pv_system_all_df["longitude"]
    )
    pv_system_all_df["x_osgb"] = x_osgb
    pv_system_all_df["y_osgb"] = y_osgb

    pv_system_all_df["capacity_kw"] = pv_system_all_df["installed_capacity_kw"]
    pv_system_all_df["capacity_watt_power"] = pv_system_all_df["capacity_kw"] * 1000

    # sort index
    pv_system_all_df = pv_system_all_df.sort_index()

    return pv_system_all_df


def get_pv_power_from_database(
    history_duration: timedelta,
    interpolate_minutes: int,
    load_extra_minutes: int,
    load_extra_minutes_and_keep: Optional[int] = 30,
    providers: List[str] = None,
) -> pd.DataFrame:
    """
    Get pv power from database

    Args:
        history_duration: a timedelta of how many minutes to load in the past
        interpolate_minutes: how many minutes we should interpolate the data froward for
        load_extra_minutes: the extra minutes we should load, in order to load more data.
            This is because some data from a site lags significantly behind 'now'.
            These extra minutes are not kept but used to interpolate results.
        load_extra_minutes_and_keep: extra minutes to load, but also keep this data.
        providers: optional list of providers

    Returns:pandas data frame with the following columns pv systems indexes
    The index is the datetime

    """

    logger.info("Loading PV data from database")
    logger.debug(f"{history_duration=} {interpolate_minutes=} {load_extra_minutes=}")

    if providers is None:
        providers = [pv_output, solar_sheffield_passiv]

    extra_duration = timedelta(minutes=load_extra_minutes)
    now = pd.to_datetime(datetime.now(tz=timezone.utc)).ceil("5T")
    start_utc = now - history_duration - timedelta(minutes=load_extra_minutes_and_keep)
    start_utc_extra = start_utc - extra_duration

    # create empty dataframe with 5 mins periods
    empty_df = pd.DataFrame(index=pd.date_range(start=start_utc_extra, end=now, freq="5T"))

    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_PV)

    with db_connection.get_session() as session:
        pv_yields: List[PVYieldSQL] = get_pv_yield(
            session=session, start_utc=start_utc_extra, correct_data=True, providers=providers
        )

        logger.debug(f"Found {len(pv_yields)} PV yields from the database")

        pv_yields_df = pd.DataFrame(
            [(PVYield.from_orm(pv_yield)).__dict__ for pv_yield in pv_yields]
        )

    if len(pv_yields_df) == 0:
        logger.warning("Found no pv yields, this might cause an error")
    else:
        logger.debug(f"Found {len(pv_yields_df)} pv yields")

    if len(pv_yields_df) == 0:

        data = create_empty_pv_data(end_utc=now, providers=providers, start_utc=start_utc)

        return data

    # get the system id from 'pv_system_id=xxxx provider=.....'
    pv_yields_df["pv_system_id"] = (
        pv_yields_df["pv_system"].astype(str).str.split(" ").str[0].str.split("=").str[-1]
    ).astype(int)

    pv_yields_df["provider"] = (
        pv_yields_df["pv_system"]
        .astype(str)
        .str.split(" ")
        .str[1]
        .str.split("=")
        .str[1]
        .str.replace("'", "")
    )

    # encode pv system id
    for provider in pv_output, solar_sheffield_passiv:
        idx = pv_yields_df["provider"] == provider

        pv_yields_df.loc[idx, "pv_system_id"] = encode_label(
            pv_yields_df.loc[idx, "pv_system_id"], label=provider
        )

    # pivot on
    pv_yields_df = pv_yields_df[["datetime_utc", "pv_system_id", "solar_generation_kw"]]
    pv_yields_df.drop_duplicates(
        ["datetime_utc", "pv_system_id", "solar_generation_kw"], keep="last", inplace=True
    )
    pv_yields_df = pv_yields_df.pivot(
        index="datetime_utc", columns="pv_system_id", values="solar_generation_kw"
    )

    # we are going interpolate using 'quadratic' method and we need at least 3 data points,
    # Lets make sure we have double that, therefore we drop system with less than 6 nans
    N = len(pv_yields_df)
    pv_yields_df = pv_yields_df.loc[:, pv_yields_df.notnull().sum() >= 6]
    logger.debug(f"Have dropped {len(pv_yields_df) - N} PV systems, as they don't have enough data")

    # interpolate in between, maximum 'live_interpolate_minutes' mins
    # note data is in 5 minutes chunks
    pv_yields_df = empty_df.join(pv_yields_df)
    limit = int(interpolate_minutes / 5)
    if limit > 0:
        try:
            pv_yields_df.interpolate(limit=limit, inplace=True, method="quadratic")
        except Exception as e:
            logger.exception(e)
            logger.debug(pv_yields_df)
            raise Exception(f"Could not do interpolate with limit {limit}")

    # filter out the extra minutes loaded
    logger.debug(f"{len(pv_yields_df)} of datetimes before filter on {start_utc}")
    pv_yields_df = pv_yields_df[pv_yields_df.index >= start_utc]
    logger.debug(f"{len(pv_yields_df)} of datetimes after filter on {start_utc}")

    return pv_yields_df


def create_empty_pv_data(
    end_utc: datetime,
    providers: List[str],
    start_utc: datetime,
    sun_elevation_limit: Optional[int] = -10,
):
    """
    The idea is to create an array of nans for pv data.

    If the sun elevation is below a given value, then the nans are filled to 0

    Args:
        end_utc: end datetime of fake data
        start_utc: start datetime of fake data
        providers: optional list of providers
        sun_elevation_limit: If there is no data, we create an array of nans.
            If the elevation is below this limit, we change the first pv system data to 0.0s.

    Returns: dataframe of pv yields
    """

    # create array of nans
    logger.debug("Adding arrange of nans to pv data")
    pv_systems = get_metadata_from_database(providers=providers)
    columns = pv_systems.index
    index = pd.date_range(start=start_utc, end=end_utc, freq="5T")
    data = pd.DataFrame(columns=columns, index=index)
    data = data.apply(pd.to_numeric)

    # For ~10 pv system, lets fill with zeros if the elevation is below {sun_elevation_limit}
    # This helps keep the right shape of data for ml.
    logger.debug(
        f"For the first 10 pv system, is the sun elevation is below {sun_elevation_limit}, "
        f"then we set pv yield values to 0."
    )
    number_pv_systems_to_fill = min(len(pv_systems.index), 10)
    logger.debug(
        f"Getting sun elevations for and datestamps {index} "
        f"for {number_pv_systems_to_fill} pv systems"
    )
    # This seems to take about 1 seconds per 100 systems
    for i in range(number_pv_systems_to_fill):
        pv_system = pv_systems.iloc[i]

        sun = calculate_azimuth_and_elevation_angle(
            latitude=pv_system.latitude, longitude=pv_system.longitude, datestamps=index
        )

        mask = sun["elevation"] < sun_elevation_limit
        data.iloc[mask, i] = 0.0
    logger.debug(f"Finished adding zeros to pv data for elevation below {sun_elevation_limit}")
    return data
